@@ -75,7 +75,7 @@
    See the docstrings of defrecord, fn, and defn for more details about how
    to use these macros."
   ;; don't exclude def because it's not a var.
-  (:refer-clojure :exclude [Keyword Symbol Inst atom defrecord defn letfn defmethod fn])
+  (:refer-clojure :exclude [Keyword Symbol Inst atom defrecord defn letfn defmethod fn MapEntry ->MapEntry])
   (:require
          [clojure.pprint :as pprint]
    [clojure.string :as str]
@@ -266,7 +266,11 @@
 
 (clojure.core/defrecord Isa [h parent]
   Schema
-  (spec [this] (leaf/leaf-spec (spec/precondition this #(isa? h % parent) #(list 'isa? % parent))))
+  (spec [this] (leaf/leaf-spec (spec/precondition this
+                                                  #(if h
+                                                     (isa? h % parent)
+                                                     (isa? % parent))
+                                                  #(list 'isa? % parent))))
   (explain [this] (list 'isa? parent)))
 
 (clojure.core/defn isa
@@ -364,7 +368,7 @@
 (def Str
   "Satisfied only by String.
    Is (pred string?) and not js/String in cljs because of keywords."
-        java.lang.String                      )
+        java.lang.String                               )
 
 (def Bool
   "Boolean true or false"
@@ -376,15 +380,15 @@
 
 (def Int
   "Any integral number"
-  (pred integer?))
+  (pred integer? 'integer?))
 
 (def Keyword
   "A keyword"
-  (pred keyword?))
+  (pred keyword? 'keyword?))
 
 (def Symbol
   "A symbol"
-  (pred symbol?))
+  (pred symbol? 'symbol?))
 
 (def Regex
   "A regular expression"
@@ -747,13 +751,17 @@
   (or (required-key? ks)
       (optional-key? ks)))
 
+(clojure.core/defn map-entry-ctor [[k v :as coll]]
+        (clojure.lang.MapEntry. k v)
+                                      )
+
 ;; A schema for a single map entry.
 (clojure.core/defrecord MapEntry [key-schema val-schema]
   Schema
   (spec [this]
     (collection/collection-spec
      spec/+no-precondition+
-     vec
+     map-entry-ctor
      [(collection/one-element true key-schema (clojure.core/fn [item-fn e] (item-fn (key e)) e))
       (collection/one-element true val-schema (clojure.core/fn [item-fn e] (item-fn (val e)) nil))]
      (clojure.core/fn [[k] [xk xv] _]
@@ -796,22 +804,26 @@
                               (mapv explain-kspec))]
       (macros/assert! (empty? duplicate-keys)
                       "Schema has multiple variants of the same explicit key: %s" duplicate-keys))
-    (concat
-     (for [[k v] (dissoc this extra-keys-schema)]
-       (let [rk (explicit-schema-key k)
-             required? (required-key? k)]
-         (collection/one-element
-          required? (map-entry (eq rk) v)
-          (clojure.core/fn [item-fn m]
-            (let [e (find m rk)]
-              (cond e (item-fn e)
-                    required? (item-fn (utils/error [rk 'missing-required-key])))
-              (if e
-                (dissoc       (if (instance? clojure.lang.PersistentStructMap m) (into {} m) m)         
-                        rk)
-                m))))))
-     (when extra-keys-schema
-       [(collection/all-elements (apply map-entry (find this extra-keys-schema)))]))))
+    (let [without-extra-keys-schema (dissoc this extra-keys-schema)]
+      (concat
+       (for [[k v] without-extra-keys-schema]
+         (let [rk (explicit-schema-key k)
+               required? (required-key? k)]
+           (collection/one-element
+            required? (map-entry (eq rk) v)
+            (clojure.core/fn [item-fn m]
+              (let [e (find m rk)]
+                (cond e (item-fn e)
+                      required? (item-fn (utils/error [rk 'missing-required-key])))
+                (if e
+                  (dissoc       (if (instance? clojure.lang.PersistentStructMap m) (into {} m) m)         
+                          rk)
+                  m))))))
+       (when extra-keys-schema
+         (let [specific-keys (set (map explicit-schema-key (keys without-extra-keys-schema)))
+               [ks vs] (find this extra-keys-schema)
+               restricted-ks (constrained ks #(not (contains? specific-keys %)))]
+           [(collection/all-elements (map-entry restricted-ks vs))]))))))
 
 (defn- map-error []
   (clojure.core/fn [_ elts extra]
@@ -1206,6 +1218,22 @@
          (try (body#) (finally (set-fn-validation! true))))
        (body#))))
 
+(def fn-validator
+  "A var that can be rebound to a function to customize the behavior
+  of fn validation. When fn validation is on and `fn-validator` is
+  bound to a function, normal argument and return value checks will
+  be substituted with a call to this function with five arguments:
+
+    direction   - :input or :output
+    fn-name     - a symbol, the function's name
+    schema      - the schema for the arglist or the return value
+    checker     - a precompiled checker to check a value against
+                  the schema
+    value       - the actual arglist or return value
+
+  The function's return value will be ignored."
+  nil)
+
 (clojure.core/defn schematize-fn
   "Attach the schema to fn f at runtime, extractable by fn-schema."
   [f schema]
@@ -1341,11 +1369,11 @@
     (cljs.core/-add-method
      ~(with-meta multifn {:tag 'cljs.core/MultiFn})
      ~dispatch-val
-     (fn ~(with-meta (gensym) (meta multifn)) ~@fn-tail))
+     (fn ~(with-meta (gensym (str (name multifn) "__")) (meta multifn)) ~@fn-tail))
     (. ~(with-meta multifn {:tag 'clojure.lang.MultiFn})
        addMethod
        ~dispatch-val
-       (fn ~(with-meta (gensym) (meta multifn)) ~@fn-tail))))
+       (fn ~(with-meta (gensym (str (name multifn) "__")) (meta multifn)) ~@fn-tail))))
 
 (defmacro letfn
   "s/letfn : s/fn :: clojure.core/letfn : clojure.core/fn"
