@@ -31,5 +31,11 @@
 (defprotocol IVectorPattern
   (split [this n]))
 
-;; marker
+(defprotocol ISyntaxTag
+  (syntax-tag [this]))
+
+;; markers
+
 (definterface IExistentialPattern)
+
+(definterface IPseudoPattern)
