@@ -6,10 +6,12 @@
   (:require [puppetlabs.puppetdb.cheshire :as json]
             [clojure.java.io]
             [clojure.core.match :as cm]
+            [clojure.set :as set]
             [clojure.tools.logging :as log]
             [clojure.walk :refer [keywordize-keys stringify-keys]]
+            [murphy :refer [try!]]
             [puppetlabs.puppetdb.query-eng :as qeng]
-            [clojure.set :as set]
+            [puppetlabs.puppetdb.query.monitor :as qmon]
             [puppetlabs.i18n.core :refer [trs tru]]
             [puppetlabs.kitchensink.core :as kitchensink]
             [schema.core :as s]
@@ -23,10 +25,13 @@
                                                       parse-order-by
                                                       parse-order-by-json]]
             [puppetlabs.puppetdb.pql :as pql]
-            [puppetlabs.puppetdb.time :refer [to-timestamp]]
-            [puppetlabs.puppetdb.utils :refer [update-when]]
+            [puppetlabs.puppetdb.time :refer [ephemeral-now-ns to-timestamp]]
+            [puppetlabs.puppetdb.utils :refer [response->channel update-when]]
             [puppetlabs.puppetdb.utils.string-formatter :refer [pprint-json-parse-exception]])
-  (:import (com.fasterxml.jackson.core JsonParseException)))
+  (:import
+   (clojure.lang ExceptionInfo)
+   (com.fasterxml.jackson.core JsonParseException)
+   (java.net HttpURLConnection)))
 
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 ;;; Schemas
@@ -46,15 +51,16 @@
    (s/optional-key :include_package_inventory) (s/maybe s/Bool)
    (s/optional-key :order_by) (s/maybe [[(s/one s/Keyword "field")
                                          (s/one (s/enum :ascending :descending) "order")]])
+   (s/optional-key :timeout) (s/maybe s/Num) ;; seconds
    (s/optional-key :origin) (s/maybe s/Str)
    (s/optional-key :distinct_resources) (s/maybe s/Bool)
    (s/optional-key :distinct_start_time) s/Any
    (s/optional-key :distinct_end_time) s/Any
    (s/optional-key :limit) (s/maybe s/Int)
+   (s/optional-key :offset) (s/maybe s/Int)
    (s/optional-key :counts_filter) s/Any
    (s/optional-key :count_by) (s/maybe s/Str)
-   (s/optional-key :summarize_by) (s/maybe s/Str)
-   (s/optional-key :offset) (s/maybe s/Int)})
+   (s/optional-key :summarize_by) (s/maybe s/Str)})
 
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 ;;; Query munging functions
@@ -244,6 +250,20 @@
     b
     (Boolean/parseBoolean b)))
 
+(defn parse-timeout [x]
+  (if (number? x)
+    x
+    (let [msg #(tru "Query timeout must be non-negative number, not {0}" x)]
+      ;; Suspect we shouldn't be throwing IllegalArgumentException,
+      ;; but that's the current approach (cf. parse-limit).
+      (when-not (string? x)
+        (throw (IllegalArgumentException. (msg))))
+      (let [n (cond
+                (re-matches #"\d+" x) (Long/parseLong x)
+                (re-matches #"\d*\.\d+" x) (Double/parseDouble x)
+                :else (throw (IllegalArgumentException. (msg))))]
+        (if (zero? n) ##Inf n)))))
+
 (pls/defn-validated validate-query-params
   "Given a set of params and a param spec, throw an error if required params
    are missing or unsupported params are present, otherwise return the params."
@@ -276,7 +296,8 @@
       (update-when [:explain] parse-explain)
       (update-when [:include_facts_expiration] coerce-to-boolean)
       (update-when [:include_package_inventory] coerce-to-boolean)
-      (update-when [:distinct_resources] coerce-to-boolean)))
+      (update-when [:distinct_resources] coerce-to-boolean)
+      (update-when [:timeout] parse-timeout)))
 
 (defn parse-json-sequence
   "Parse a query string as JSON. Parse errors will result in an
@@ -318,7 +339,7 @@
       (log/warn (trs "Parsing PQL took {0} ms for PDBQuery:{1}:{2}" elapsed query-uuid (pr-str query))))
     result))
 
-(defn parse-json-or-pql-to-ast
+(defn parse-json-or-pql-query
   "Parse a query string either as JSON or PQL to transform it to AST"
   [query query-uuid log-queries?]
   (when query
@@ -367,28 +388,91 @@
      (throw (IllegalArgumentException. (tru "PuppetDB queries must be made via GET/POST"))))
    param-spec))
 
-(defn extract-query
-  "Query handler that converts the incoming request (GET or POST)
-  parameters/body to a pdb query map"
+(defn wrap-typical-query
+  "Wrap a query endpoint request handler with the normal behaviors.
+  Augment the query handler with code to extract the query from the
+  incoming GET or POST and include it in the request as a pdb query
+  map, and handle query timeouts."
   ([handler param-spec]
-   (extract-query handler param-spec parse-json-query))
+   (wrap-typical-query handler param-spec parse-json-query))
   ([handler param-spec parse-fn]
-   (fn [{:keys [puppetdb-query] :as req}]
-     (handler
-      (if puppetdb-query
-        req
-        (let [query-uuid (str (java.util.UUID/randomUUID))
-              req-with-query-uuid (assoc req :puppetdb-query-uuid query-uuid)
-              query-map (create-query-map req-with-query-uuid param-spec parse-fn)
-              pretty-print (:pretty query-map
-                                    (get-in req [:globals :pretty-print]))]
-          (-> req-with-query-uuid
-              (assoc :puppetdb-query query-map)
-              (assoc-in [:globals :pretty-print] pretty-print))))))))
+   (fn handle-typical-query [{:keys [globals puppetdb-query] :as req}]
+     (if puppetdb-query ;; apparently we may nest calls...
+       (handler req)
+       (let [{:keys [query-monitor]} globals
+             [new-req mon-id]
+             (try!
+               (let [start-ns (ephemeral-now-ns) ;; capture at the top
+                     query-uuid (str (java.util.UUID/randomUUID))
 
-(defn extract-query-pql
-  [handler param-spec]
-  (extract-query handler param-spec parse-json-or-pql-to-ast))
+                     {:keys [pretty-print query-timeout-default query-timeout-max scf-read-db]}
+                     globals
+
+                     req-with-query-uuid (assoc req :puppetdb-query-uuid query-uuid)
+                     query-map (create-query-map req-with-query-uuid param-spec parse-fn)
+                     ;; Right now, query parsing (above) has no timeouts.
+                     ;; sync is expected to override (based on its own deadlines)
+                     sync-timeout (when (#{"puppet:puppetdb-sync-batch"
+                                           "puppet:puppetdb-sync-summary"}
+                                         (:origin query-map))
+                                    (:timeout query-map query-timeout-default))
+                     timeout (or sync-timeout
+                                 (min (:timeout query-map query-timeout-default)
+                                      query-timeout-max))
+                     deadline (+ start-ns (* timeout 1000000000))
+                     ;; Wait one extra second for time-limited-seq and
+                     ;; statement timeouts since the monitor kills the
+                     ;; entire pg worker.
+                     monitor-deadline (+ deadline 1000000000)
+
+                     ;; May have no response because some tests (e.g. some
+                     ;; with-http-app based tests) don't add one right now.
+                     monitor-id (when-let [chan (and query-monitor
+                                                     (some-> (:response req)
+                                                             response->channel))]
+                                  (qmon/stop-query-at-deadline-or-disconnect query-monitor
+                                                                             query-uuid
+                                                                             chan
+                                                                             monitor-deadline
+                                                                             scf-read-db))]
+                 (try!
+                   [(-> req-with-query-uuid
+                        (assoc :puppetdb-query query-map)
+                        (update :globals merge
+                                {:pretty-print (:pretty query-map pretty-print)
+                                 :query-deadline-ns deadline
+                                 :query-monitor query-monitor}
+                                (when monitor-id
+                                  {:query-monitor-id monitor-id})))
+                    monitor-id]
+                   (catch Exception ex
+                     (when monitor-id (qmon/forget query-monitor monitor-id))
+                     (throw ex))))
+               (catch ExceptionInfo ex
+                 (when-not (= :puppetlabs.puppetdb.query/timeout (:kind (ex-data ex)))
+                   (throw ex))
+                 ;; Note, this will not be reached for the typical
+                 ;; streaming query case because after the query starts, we
+                 ;; return from this function with a 200, and the stream,
+                 ;; and further errors/timeouts have to be handled by the
+                 ;; generated-stream thread feeding jetty.
+                 (let [{:keys [id origin]} (ex-data ex)]
+                   (log/warn (.getMessage ex))
+                   (http/error-response
+                    (if origin
+                      (tru "Query {0} from {1} exceeded timeout" id (pr-str origin))
+                      (tru "Query {0} exceeded timeout" id))
+                    HttpURLConnection/HTTP_INTERNAL_ERROR))))]
+         (if-not mon-id
+           (handler new-req)
+           (try!
+             (let [res (handler new-req)]
+               (when-not (:puppetlabs.puppetdb.query/streaming res)
+                 (qmon/forget query-monitor mon-id))
+               res)
+             (catch Throwable ex
+               (qmon/forget query-monitor mon-id)
+               (throw ex)))))))))
 
 (defn validate-distinct-options!
   "Validate the HTTP query params related to a `distinct_resources` query.  Return a
@@ -430,7 +514,8 @@
   [globals]
   (select-keys globals [:scf-read-db :warn-experimental :url-prefix
                         :pretty-print :node-purge-ttl :add-agent-report-filter
-                        :log-queries]))
+                        :log-queries :query-deadline-ns :query-monitor
+                        :query-monitor-id :puppetlabs.puppetdb.config/test]))
 
 (defn valid-query?
   [scf-read-db version query-map query-options]
