@@ -5,12 +5,13 @@
             [clojure.tools.logging :as log]
             [honey.sql :as hsql]
             [puppetlabs.kitchensink.core :as kitchensink]
-            [puppetlabs.puppetdb.time :as pl-time]
+            [puppetlabs.puppetdb.time :as pl-time :refer [ephemeral-now-ns]]
             [puppetlabs.puppetdb.jdbc.internal :refer [limit-result-set!]]
             [puppetlabs.puppetdb.schema :as pls :refer [defn-validated]]
             [schema.core :as s]
             [puppetlabs.i18n.core :refer [trs]])
   (:import
+   (clojure.lang ExceptionInfo)
    (com.zaxxer.hikari HikariDataSource HikariConfig)
    (java.sql Connection SQLException SQLTransientConnectionException)
    (org.postgresql.util PGobject)))
@@ -22,14 +23,23 @@
 
 (defn sql-state [kw-name]
   (or ({:admin-shutdown "57P01"
+        :idle-in-transaction-session-timeout "25P03"
+        :idle-session-timeout "57P05"
         :invalid-regular-expression "2201B"
-        :program-limit-exceeded "54000"
         :lock-not-available "55P03"
+        :program-limit-exceeded "54000"
         :query-canceled "57014"}
        kw-name)
       (throw (IllegalArgumentException.
               (trs "Requested unknown SQL state")))))
 
+(defn clj-jdbc-termination-ex? [ex]
+  (and (instance? ExceptionInfo ex)
+       (let [{:keys [rollback handling]} (ex-data ex)]
+         (and rollback
+              (= (sql-state :admin-shutdown)
+                 (some-> handling .getSQLState))))))
+
 (defmacro with-db-connection [spec & body]
   `(sql/with-db-connection [db# ~spec]
      (binding [*db* db#]
@@ -85,9 +95,11 @@
         orig (.getAutoCommit conn)]
     (.setAutoCommit conn true)
     (try
-      (doseq [c commands]
+      (doseq [x commands
+              :let [batch (if (string? x) [x] x)]
+              cmd batch]
         (with-open [s (.createStatement conn)]
-          (.execute s c)))
+          (.execute s cmd)))
       (finally (.setAutoCommit conn orig)))))
 
 (defn insert!
@@ -303,12 +315,15 @@
   call-with-query-rows, and apply any necessary conversions directly
   to each row.  (Most columns cannot be arrays.)"
   ([query f] (call-with-array-converted-query-rows query {} f))
-  ([[sql & params] {:keys [as-arrays?] :as opts} f]
+  ([[sql & params]
+    {:keys [as-arrays? fetch-size]
+     :or {fetch-size 500} :as opts}
+    f]
    (with-db-transaction []
      (with-open [stmt (.prepareStatement ^Connection (:connection *db*) sql)]
        (doseq [[i param] (map vector (range) params)]
          (.setObject stmt (inc i) param))
-       (.setFetchSize stmt 500)
+       (.setFetchSize stmt fetch-size)
        (let [fix-vals (if as-arrays?
                         #(mapv any-sql-array->vec %)
                         #(kitchensink/mapvals any-sql-array->vec %))]
@@ -709,6 +724,9 @@
 (defn current-database []
   (-> "select current_database();" query-to-vec first :current_database))
 
+(defn current-pid []
+  (-> "select pg_backend_pid();" query-to-vec first :pg_backend_pid))
+
 (defn-validated has-database-privilege? :- s/Bool
   [user db privilege]
   (-> ["select has_database_privilege(?, ?, ?)" user db privilege]
@@ -756,3 +774,32 @@
   (do-commands
    (format "grant connect on database %s to %s"
           (double-quote db) (double-quote role))))
+
+(defn local-timeout-ex?
+  "Returns true if ex is an exception that might have been thrown as a
+  result of the timeouts set by update-local-timeouts."
+  [ex]
+  (let [state (.getSQLState ex)]
+    (case state ;; https://www.postgresql.org/docs/current/errcodes-appendix.html
+      ;; query_canceled
+      ;; idle_in_transaction_session_timeout
+      ("57014" "25P03") true
+      false)))
+
+(defn- ms-until-ns-deadline
+  [deadline-ns]
+  (int (quot (- deadline-ns (ephemeral-now-ns))
+             1000000)))
+
+(defn update-local-timeouts
+  "Sets the local timeouts (idle and statement) to respect
+  deadline-ns unless the deadline has passed, then sets them to
+  min-ms."
+  [deadline-ns min-ms]
+  (assert (pos? min-ms))
+  (when (some-> deadline-ns (not= ##Inf))
+    (let [timeout-ms (ms-until-ns-deadline deadline-ns)
+          timeout-ms (if (pos? timeout-ms) timeout-ms min-ms)]
+      (do-commands
+       (format "set local statement_timeout = %d" timeout-ms)
+       (format "set local idle_in_transaction_session_timeout = %d" timeout-ms)))))
